import io
import os
import shutil
from typing import Tuple
import zipfile

from multipledispatch import dispatch

from ..utils import removeprefix
from .http import HTTPTransport


class GithubTransport(HTTPTransport):
    """Transport for Github repository.

    This transport check specified URI has valid Github repository URL. (e.g. https://github.com/)
    """

    schemes = ("https://github.com/", "http://github.com/")

    @staticmethod
    def is_supported(uri: str) -> bool:
        return any(uri.startswith(prefix) for prefix in GithubTransport.schemes)

    @dispatch(str, str)
    async def read(self, uri: str, path: str) -> bytes:
        """Read a file binary data from the specified URI and path.

        Note that Github downloadable URL is different from the URI itself. So We are replacing the
        URI and path into the valid URL and calling `read(location)` of `HTTPTransport` here.
        """
        assert self.is_supported(uri)

        owner, repo, branch = self.parse(uri)
        return await self.read(f"https://raw.githubusercontent.com/{owner}/{repo}/{branch}/{path}")

    @dispatch(str)  # type: ignore
    async def read(self, location: str) -> bytes:  # noqa: F811
        return await super().read(location)

    def parse(self, uri: str) -> Tuple[str, str, str]:
        """Parse URI following Github repository scheme.

        e.g. https://github.com/furiosa-ai/furiosa-models:main to (furiosa-ai, furiosa-models, main)
        """

        # Remove Github URI prefix
        for prefix in self.schemes:
            uri = removeprefix(uri, prefix)

        # Default branch is "main"
        branch = "main"
        owner, repo = uri.split("/")
        if ":" in repo:
            repo, branch = repo.split(":")

        return owner, repo, branch

    async def download(self, uri: str):
        def unzip(data: bytes, where: str) -> str:
            with zipfile.ZipFile(io.BytesIO(data)) as f:
                name = f.infolist()[0].filename

                target = os.path.join(where, name)

                shutil.rmtree(target, ignore_errors=True)

                f.extractall(where)
            return target

        # TODO(ileixe): Impmlement cache layer.
        assert self.is_supported(uri)

        directory = self.cache_directory
        if not os.path.exists(directory):
            os.makedirs(directory)

        owner, repo, branch = self.parse(uri)

        data = await self.read(f"https://github.com/{owner}/{repo}/archive/{branch}.zip")

        src = unzip(data, directory)
        dst = os.path.join(directory, f"{owner}_{repo}_{branch}")

        shutil.rmtree(dst, ignore_errors=True)
        shutil.move(src, dst)

        return dst



# TODO: Comment out S3 tests as moto does not support aiobotocore. I don't want to make mock
# server just to verify basic S3 API, so wait for pytest plugin now.

# import boto3
# import pytest
# from furiosa.comon.transport import S3Transport
# from moto import mock_s3


# @pytest.fixture
# def bucket():
#     return "BUCKET_NAME"


# @pytest.fixture(autouse=True)
# @mock_s3
# def prepare(artifact_file, model_file, bucket):
#     """
#     Prepare mock s3 server fixture to serve files.
#     """
#     client = boto3.client("s3")

#     with open(model_file, "rb") as data:
#         client.upload_fileobj(data, bucket, artifact_file)

#     with open(artifact_file, "rb") as data:
#         client.upload_fileobj(data, bucket, model_file)

#     yield


# @pytest.fixture(scope="function")
# def transport() -> S3Transport:
#     return S3Transport()


# @pytest.mark.asyncio
# async def test_list(transport, bucket):
#     assert (await transport.list(f"s3://{bucket}/")) == ["MNISTNet"]


# @pytest.mark.asyncio
# async def test_download(transport, bucket, model_file, model_binary):
#     assert await transport.download(f"s3://{bucket}/{model_file}") == model_binary



"""
Common fixtures to be used from all tests.
"""

import pytest


@pytest.fixture(scope="module")
def model_file() -> str:
    return "./tests/unit/fixtures/MNISTnet_uint8_quant_without_softmax.tflite"


@pytest.fixture(scope="module")
def model_binary(model_file) -> bytes:
    with open(model_file, "rb") as data:
        return data.read()



"""Transport utilities to fetch files from various sources."""

from contextlib import contextmanager
from typing import Iterator

from multipledispatch import dispatch

from .base import Transport
from .error import TransportNotFound
from .file import FileTransport
from .github import GithubTransport
from .http import HTTPTransport
from .s3 import S3Transport

__all__ = [
    "transports",
    "FileTransport",
    "GithubTransport",
    "HTTPTransport",
    "S3Transport",
    "Transport",
    "supported",
    "read",
    "download",
    "is_relative",
]

transports = [FileTransport(), GithubTransport(), HTTPTransport(), S3Transport()]


@contextmanager
def supported(uri: str) -> Iterator[Transport]:
    """Supported transport for the URI."""
    for transport in transports:
        if transport.is_supported(uri):
            # Note that each transport should be exclusive as we are returning first one.
            # TODO(ileixe): Add more logging to clarify what's going on here to user.
            yield transport
            return

    raise TransportNotFound(uri, transports)


@dispatch(str, str)
async def read(uri: str, path: str) -> bytes:
    """Read a file binary data from the URI and path with a transport which supports the URI.

    Args:
        uri (str): URI to locate the models.
        path (str): Relative file path in the repositry to read.

    Returns:
        bytes: Downloaded binary data.

    Raises:
        TransportNotFound: If all the available transports are not supporting the URI.
    """
    with supported(uri) as transport:
        return await transport.read(uri, path)


@dispatch(str)  # type: ignore
async def read(location: str) -> bytes:  # noqa: F811
    """Read a file binary data from the specified location with a transport which supports the URI.

    Args:
        location (str): Location(URL) to read the file. This should be valid URL to download.

    Returns:
        bytes: Downloaded binary data.

    Raises:
        TransportNotFound: If all of the available transports are not supporing the URI.
    """
    with supported(location) as transport:
        return await transport.read(location)


async def download(uri: str) -> str:
    """Download a directory into local destination with a transport which supports the URI.

    Args:
        uri (str): URI to download the data.

    Returns:
        str: Destination directory name. This directory will be located in `cache` directory.

    Raises:
        TransportNotFound: If all the available transports are not supporting the URI.
    """
    with supported(uri) as transport:
        return await transport.download(uri)


def is_relative(path: str) -> bool:
    """Is this path relative path?

    If all the available transports are not supporting the path, we assume that it's a relative
    path. You should find the path from the URI if it's a relative path.
    """
    return all(not transport.is_supported(path) for transport in transports)



import io
from typing import Tuple
from urllib.parse import urlparse

from multipledispatch import dispatch

from .base import Transport


class S3Transport(Transport):
    """Transport for S3.

    This transport check specified URI has valid S3 scheme (e.g. s3://).
    """

    scheme = "s3://"

    @staticmethod
    def is_supported(uri: str) -> bool:
        return uri.startswith(S3Transport.scheme)

    @dispatch(str, str)
    async def read(self, uri: str, path: str) -> bytes:
        assert self.is_supported(uri)
        return await self.read(f"{uri}/{path}")

    @dispatch(str)  # type: ignore
    async def read(self, location: str) -> bytes:  # noqa: F811
        import aioboto3

        async with aioboto3.Session().resource("s3") as resource:
            bucket, key = self.parse(location)

            with io.BytesIO() as data:
                await (await resource.Object(bucket, key)).download_fileobj(data)
                return data.getvalue()

    async def download(self, uri: str) -> str:
        raise NotImplementedError("S3Transport download() not yet supported")

    @staticmethod
    def parse(uri: str) -> Tuple[str, str]:
        """Parse URI to network location and path."""
        parsed = urlparse(uri)
        return parsed.netloc, parsed.path[1:]



import io
from typing import Dict

import aiohttp
from multipledispatch import dispatch
from tqdm.asyncio import tqdm

from .base import Transport


class HTTPTransport(Transport):
    """Transport for HTTP.

    This transport check specified URI has valid http scheme (e.g. https://, http://)
    """

    schemes = ("https://", "http://")

    def __init__(self, headers: Dict = {}, params: Dict = {}):
        self._headers = headers
        self._params = params

    @staticmethod
    def is_supported(uri: str) -> bool:
        return any(uri.startswith(scheme) for scheme in HTTPTransport.schemes)

    @dispatch(str, str)
    async def read(self, uri: str, path: str) -> bytes:
        assert self.is_supported(uri)

        return await self.read(f"{uri}/{path}")

    @dispatch(str)  # type: ignore
    async def read(self, location: str) -> bytes:  # noqa: F811
        async with aiohttp.ClientSession(headers=self._headers) as session:
            async with session.get(location, params=self._params) as response:
                with io.BytesIO() as data:
                    response.raise_for_status()

                    total = int(response.headers.get("content-length", 0))

                    async for chunk in tqdm(
                        response.content.iter_chunked(1024), total=(total + 1024) // 1024
                    ):
                        data.write(chunk)

                    return data.getvalue()

    async def download(self, uri: str) -> str:
        raise NotImplementedError("HTTPTransport download() not yet supported")



from typing import TYPE_CHECKING, List

if TYPE_CHECKING:
    from .base import Transport


class TransportError(Exception):
    def __init__(self, msg):
        super().__init__(msg)


class TransportNotFound(TransportError):
    def __init__(self, uri: str, transports: List["Transport"]):
        msg = f"Transport for {uri} not found. Available transports:\n\n"

        msg += "\n".join(
            f"{type(transport).__name__}: {transport.__doc__}" for transport in transports
        )

        super().__init__(msg)



from aiohttp import web
import pytest

from furiosa.common import transport


@pytest.fixture(scope="function")
async def server(aiohttp_server):
    """
    Running aiohttp application server fixture to serve static files.
    """
    app = web.Application()
    app.router.add_routes([web.static("/", ".")])

    # Run AioHttp test server.
    yield await aiohttp_server(app)


@pytest.mark.asyncio
async def test_read(server, model_file, model_binary):
    with transport.supported("http://") as client:
        assert await client.read(f"http://{server.host}:{server.port}/{model_file}") == model_binary



import os
import pathlib
import shutil

import aiofiles
from multipledispatch import dispatch

from ..utils import removeprefix
from .base import Transport


class FileTransport(Transport):
    """Transport for local file path.

    This transport check specified URI has valid local file scheme (e.g. file://).
    """

    scheme = "file://"

    @staticmethod
    def is_supported(uri: str) -> bool:
        return uri.startswith(FileTransport.scheme)

    @dispatch(str, str)
    async def read(self, uri: str, path: str) -> bytes:
        assert self.is_supported(uri)

        return await self.read(os.path.join(removeprefix(uri, self.scheme), path))

    @dispatch(str)  # type: ignore
    async def read(self, location: str) -> bytes:  # noqa: F811
        file = removeprefix(location, self.scheme)
        async with aiofiles.open(file, "rb") as f:
            return await f.read()

    async def download(self, uri: str) -> str:
        """Download a directory into local destination.

        FileTransport download is just to copy local directory into the cache.
        """
        # TODO(ileixe): Implement cache layer.
        assert self.is_supported(uri)

        directory = self.cache_directory
        if not os.path.exists(directory):
            os.makedirs(directory)

        location = removeprefix(uri, self.scheme)

        src = location
        dst = os.path.join(directory, pathlib.Path(location).absolute().name)

        shutil.rmtree(dst, ignore_errors=True)
        return shutil.copytree(src, dst)



from abc import ABC, abstractmethod
import os

from multipledispatch import dispatch


class Transport(ABC):
    """Base transport class to read/download from various sources."""

    @staticmethod
    def is_supported(uri: str) -> bool:
        """Decide whether this transport supports the specified URI.

        Args:
            uri (str): URI to locate the models.

        Returns:
            bool: This transport supports the URI or not.
        """
        raise NotImplementedError

    @dispatch(str, str)
    async def read(self, uri: str, path: str) -> bytes:
        """Read a file binary data from the specified URI and path.

        This is a high level function to use `read(location)` internally.

        Args:
            uri (str): URI to locate the models.
            path (str): Relative file path in the repositry to read.

        Returns:
            bytes: Downloaded binary data.
        """
        raise NotImplementedError

    @dispatch(str)  # type: ignore
    async def read(self, location: str) -> bytes:  # noqa: F811
        """Read a file binary data from the specified location.

        Args:
            location (str): Location(URL) to read the file. This should be valid URL to download.

        Returns:
            bytes: Downloaded binary data.
        """
        raise NotImplementedError

    @abstractmethod
    async def download(self, uri: str) -> str:
        """Download a directory into local destination.

        Args:
            uri (str): URI to download the data.

        Returns:
            str: Destination directory name. This directory will be located in `cache` directory.
        """
        raise NotImplementedError

    @property
    def cache_directory(self) -> str:
        """Cache directory to save downloaded files."""
        return os.path.expanduser(
            os.getenv(
                "FURIOSA_TRANSPORT_HOME",
                os.path.join(os.getenv("XDG_CACHE_HOME", "~/.cache"), "furiosa"),
            )
        )



import pytest

from furiosa.common import transport


@pytest.mark.asyncio
async def test_read(model_file, model_binary):
    with transport.supported("file://") as client:
        assert await client.read(model_file) == model_binary
